
#tarjar
#Julian Marcel Noble 2005. BSD License. [JMN] on the Tcler's wiki - http://wiki.tcl.tk
#
#------------------------------------------------------------------------------------
#
# Purpose: Read (load) & write tarjarred module files using pure-tcl.
#   This might be useful as a bootstrap for other (faster,smaller,funkier?) module-archivers that use binaries. 
#   e.g trofs or starkit ??
#
#    A tarjar is a valid tarball with the contents under a specially named folder. 
#    This foldername is a Tcl commented string: #tarjar-<package>-<version>
#    The first file in the archive starts with a linefeed, ends with a comment character (#) and contains the load-script.
#	 There is also a file named #z that ends with a ctrl-z character to terminate Tcl sourcing.
#
# A tarjar .tm file is sourceable by Tcl and thus directly loadable as a Tcl module.
# The tarjar format offers the advantage that the contents are accessible using 
# standard tar archive tools as well as other commandline tools such as grep.
# 
#
# This package only does a 'package require tar' (tcllib) if creating a tarjar.
# Packages tar, vfs & any package requiring binaries, must not be required for reading.
# Preferably this package will not require any other packages. 
# - It'd be nice to keep things ultra simple for relatively easy deployment by people unfamiliar with Tcl packages & modules.
#   
#
# !todo - investigate: Is there an *efficient* way to test for and utilize a binary tar package that may be present?
#  (i.e a method that doesn't trigger scans of the auto_path, or other extensive filesystem searches.)
#  Will this offer any worthwhile performance benefits, or should other module-archivers be used instead if that is an issue?
#------------------------------------------------------------------------------------
#
# zipped-tarjars
# With thanks to Pascal Scheffers 
# see Using zip files like a .kit file: http://wiki.tcl.tk/13876
#
#------------------------------------------------------------------------------------


#!todo - don't maintain connection to .tm file after diversion to unwrapped version.
# (so how can we deliberately connect to the .tm file in this case?)

#fix tarjar::create - works as needed by tarjar::wrap - but doesn't properly pack all supplied files when called directly.
#------------------------------------------------------------------------------------



#don't do a 'package provide' until we've given a chance to redirect..

#Changes:
#version 2.3 (2023-03)
#   avoid any module/lib loading for the tarjar::connect side of tarjar operations (startup-time reduction and dependency minimization)
#   - inline the very few struct::set functions required (tcllib 1.20)
#   - inline the platform module code distributed with tcl, as it's still a separate library reducing the ability to run with reduced lib/module scanning during startup (platform 1.0.17)
#   - see #maintenance-note tags for versions to update as necessary
#  add -name option to tarjar::wrap so that modulename not necessarily taken from name of .tcl script being wrapped.
#Version 1.5
#	Source all scripts with -encoding utf-8



namespace eval tarjar {
	variable LogLevel
	if {![info exists LogLevel]} {
		# var may exist if namespace was explicitly created in order to pre-set the loglevel!
		set LogLevel warn
	}

	if {![info exists ::tarjar::debug]} {
		set debug 0     ;#this does not affect LogLevel! 
	}

    variable connected 
    if {![info exists connected(to)]} {
        set connected(to) [list]
    }

	set ::tarjar::tarjarscript [info script]
	
	if {[string tolower [file extension $::tarjar::tarjarscript]] eq ".tcl"} {
		set connected(self) [file dirname $::tarjar::tarjarscript]
	} else {
		#presumably a .tm
		set connected(self) $tarjarscript
	}
	
    #extensions of loadable files
    #variable loadables [list .so .dll]
    variable loadables [info sharedlibextension]
	variable sourceables {.tcl .tk}    ;#.tm? 

    variable version
}

#package require struct::set  ;#faster than package require struct (we don't need whole package)
#BUT - the dependency on tcllib was causing load-time issues - especially with tcllibc (present or absent - both search and load add to startup time)
#for the trivial purposes of tarjar - the tcl versions are fine, and so the few required functions are borrowed and inlined here under a different namespace
#maintenance-note: tcllib 1.20
namespace eval tarjar::structset {
    #some direct borrows from tcllib struct::set
    proc intersect {A B} {
        if {[llength $A] == 0} {return {}}
        if {[llength $B] == 0} {return {}}

        # This is slower than local vars, but more robust
        if {[llength $B] > [llength $A]} {
        ::set res $A
        ::set A $B
        ::set B $res
        }
        ::set res {}
        foreach x $A {::set ($x) {}}
        foreach x $B {
        if {[info exists ($x)]} {
            lappend res $x
        }
        }
        return $res
    }
    # ::struct::set::S_difference --
    #
    #	Compute difference of two sets.
    #
    # Parameters:
    #	A, B	-- Sets to compute the difference for.
    #
    # Results:
    #	A - B
    #
    # Side effects:
    #       None.

    proc difference {A B} {
        if {[llength $A] == 0} {return {}}
        if {[llength $B] == 0} {return $A}

        array set tmp {}
        foreach x $A {::set tmp($x) .}
        foreach x $B {catch {unset tmp($x)}}
        return [array names tmp]
    }
}


set tarjar::hasLogger false

if 0 {
if {![catch {package require logger} logger_version]} {
    set tarjar::hasLogger true
    set tarjar::logger_version $logger_version
}
}



namespace eval tarjar::_ {
    #namespace for internal procs
    variable last_scanned_tmlist 
    if {![info exists last_scanned_tmlist]} {
        set last_scanned_tmlist [list]
    }
}

#As tarjar is likely to be loaded very early in a script's life - 
# we must be careful about writing to stdout/stderr.
# use ::tarjar::Puts in this package instead of puts to emit warnings.
# stderr/stdout may have been redirected or not yet be open for writing.
# (e.g script might require other packages before standard channels are in a usable state)
# 
#   We could just wrap all warnings in a catch... but then warnings are lost.
#

#tarjar 'cooperates' with the channelguard package since we can't load it from within this package.
# (i.e there is a tighter than 'normal' coupling between this package and channelguard - they should be updated in tandem.. or at least have their common assumptions reviewed.)
namespace eval ::channelguard {
    variable buffer      ;#note - this will not cause it to 'exist' as far as '[info exists...]' is concerned.
    if {![::info exists buffer]} {
	    array set buffer [list] ;#make sure it exists so we can append to $chan,data elements.
    }

}
proc ::tarjar::Puts {args} {
    set nl \n
    if {[llength $args] > 1 && ([lindex $args 0] eq "-nonewline")} {
        set nl ""
        set args [lrange $args 1 end]
    }
    if {[llength $args] == 1} {
        #default to stdout
        set args [list stdout [join $args]]
    }

    lassign $args channel s

    set cmd ::puts
    if {$nl==""} {
        lappend cmd -nonewline
    }

    if {![info exists ::channelguard::README-tarjar]} {
        set ::channelguard::README-tarjar {
The ::channelguard namespace belongs to the 'channelguard' package.
The 'tarjar' package may have written directly to the channelguard::buffer variable. 
        }
    }

	if {[catch {puts -nonewline $channel ""}]} {
		#channel not writeable
		#so where do we send our data??		
		#for now - just accumulate in a variable because loading other packages such as Memchan may also invoke puts. 
		# (depending on package distribution method etc)
		#!todo - review
		#   append ::tarjar::$channel $s

        append ::channelguard::buffer($channel,data) $s$nl
	} else {
        #{*}$cmd $channel $s


        
        if {![catch {package present channelguard}]} {
			#once we know channelguard is loaded - we can replace tarjar::Puts with the standard puts 
            rename ::tarjar::Puts {}
            interp alias {} ::tarjar::Puts {} ::puts
            {*}$cmd $channel $s
            return

            #review. guarded or not. Channelguard is loaded.. so shouldn't we just alias our Puts to puts now?
            #e.g interp alias {} tarjar::Puts {} ::puts
        	if {$channel ni [::channelguard::guarded]} {
        		#channelguard loaded but this channel not guarded..
        		#presumably therefore it's now ok to write to it.
        		{*}$cmd $channel $s
        	} else {
                #!todo - review. should we really be writing direct to channelguard buffer here.. or just  let channelguard handle it??
        		append ::channelguard::buffer($channel,data) $s$nl
        	}
        } else {
        	append ::channelguard::buffer($channel,data) $s$nl
        }
	}
}


namespace eval tarjar {
	variable logger
	variable hasLogger
	if {$hasLogger} {
		set logger [logger::init tarjar]
		${logger}::setlevel $::tarjar::LogLevel
		
		proc ::tarjar::LogLevel {args} {
			if {![llength $args]} {
				return [${tarjar::logger}::currentloglevel]
			} else {
				${tarjar::logger}::setlevel $args
				set ::tarjar::LogLevel $args  ;#keep this in sync too.. 
			}
		}
		
		proc ${logger}::stdoutcmd {level text} {
			set logline "[clock format [clock seconds] -format  %Y-%m-%dT%H:%M:%S] tarjar $level $text"
			if {$level in {error critical}} {
				puts stderr $logline
			} else {
				puts stdout $logline
			}
		}
		
		proc ::tarjar::Log {level text} {
			variable logger
			${logger}::${level} $text
		}
		
	} else {
		proc ::tarjar::Log {args} {}
	}
	
}


proc ::tarjar::myversion {} {
	set script [info script]
	if {![string length $script]} {
		error "No result from \[info script\] - Function should only be called from within a loading tarjar"
	}	


    set fname [file tail [file rootname [file normalize $script]]]
	set scriptdir [file dirname $script]
	
    if {![string match "#tarjar-*" $fname]} {
       lassign [lrange [split $fname -] end-1 end] _pkgname version
    } else {
        lassign [scan [file tail [file rootname $script]] {#tarjar-loadscript-%[a-z]-%s}] _pkgname version
        if {![string length $version]} {
            #try again on the name of the containing folder
            lassign [scan [file tail $scriptdir] {#tarjar-%[a-z]-%s}] _pkgname version
			#todo - proper walk up the directory tree
			if {![string length $version]} {
				#try again on the grandparent folder (this is a standard depth for sourced .tcl files in a tarjar)
				lassign [scan [file tail [file dirname $scriptdir]] {#tarjar-%[a-z]-%s}] _pkgname version
			}
        }
    }

	tarjar::Log debug "'myversion' determined version for [info script]:  $version"
	return $version
}


proc ::tarjar::myname {} {
    set script [info script]
    if {![string length $script]} {
        error "No result from \[info script\] - Function should only be called from within a loading tarjar"
    }
    #return $::tarjar::connected(package,[string tolower $script])
    return $::tarjar::connected(package,$script)
}
proc ::tarjar::myfullname {} {
    set script [info script]
    set script [::tarjar::normalize $script] 
    if {![string length $script]} {
        error "No result from \[info script\] - Function should only be called from within a loading tarjar"
    }
    return $::tarjar::connected(fullpackage,$script)
}



proc ::tarjar::normalize {path} {
	#Tcl's 'file normalize' seems to do some unfortunate tilde substitution on windows.. (at least for relative paths)
	# we take the assumption here that if Tcl's tilde substitution is required - it should be done before the path is provided to this function.
	set matilda "<_tarjar_tilde_placeholder_>"  ;#token that is *unlikely* to occur in the wild, and is somewhat self describing in case it somehow ..escapes.. 
	set path [string map [list ~ $matilda] $path]  ;#give our tildes to matilda to look after
	set path [file normalize $path]
    #set path [string tolower $path] ;#must do this after file normalize
	return [string map [list $matilda ~] $path]  ;#get our tildes back.
}


#------------------------------------------------------------
#see returneval: http://wiki.tcl.tk/4346  

proc ::tarjar::_::returneval {script} {return -code -1 $script}

proc ::tarjar::_::eproc {name arglist body} {
	 set alt ::tarjar::_::[namespace tail $name]

     uplevel 1 [list proc $alt $arglist $body]
     interp alias {} $name {} [namespace which -command ::tarjar::_::eproc_call] $alt
}
proc ::tarjar::_::eproc_call {args} {
     set code [catch [list uplevel 1 $args] res]
     if {$code == -1} {
         set code [catch [list uplevel 1 $res] res]
         return -code $code $res
     } elseif {$code == 1} {
         return -code error -errorinfo $::errorInfo -errorcode $::errorCode $res
     } else {
         return -code $code $res
     }
}
#------------------------------------------------------------



#from tcllib tar
proc ::tarjar::_::pad {size} {
    set pad [expr {512 - ($size % 512)}]
    if {$pad == 512} {return 0}
    return $pad
}

#straight snaffle from from tar::readHeader in tcllib 1.19!
#tcllib tar API appears only designed to deal with entire tarball files, not streams, otherwise 
# we would have simply used that tar package.
#
proc ::tarjar::_::readHeader1 {data} {
    binary scan $data a100a8a8a8a12a12a8a1a100a6a2a32a32a8a8a155 \
                      name mode uid gid size mtime cksum type \
                      linkname magic version uname gname devmajor devminor prefix
                               
    foreach x {name mode type linkname magic uname gname prefix mode uid gid size mtime cksum version devmajor devminor} {
        set $x [string trim [set $x] "\x00"]
    }
    set mode [string trim $mode " \x00"]
    foreach x {uid gid size mtime cksum version devmajor devminor} {
        set $x [format %d 0[string trim [set $x] " \x00"]]
    }

    return [list name $name mode $mode uid $uid gid $gid size $size mtime $mtime \
                 cksum $cksum type $type linkname $linkname magic $magic \
                 version $version uname $uname gname $gname devmajor $devmajor \
                 devminor $devminor prefix $prefix]
}

proc ::tarjar::_::readHeader {data} {
    binary scan $data a100a8a8a8a12a12a8a1a100a6a2a32a32a8a8a155 \
                      name mode uid gid size mtime cksum type \
                      linkname magic version uname gname devmajor devminor prefix

    foreach x {name type linkname} {
        set $x [string trim [set $x] "\x00"]
    }
    foreach x {uid gid size mtime cksum} {
        set $x [format %d 0[string trim [set $x] " \x00"]]
    }
    set mode [string trim $mode " \x00"]

    if {$magic == "ustar "} {
        # gnu tar
        # not fully supported
        foreach x {uname gname prefix} {
            set $x [string trim [set $x] "\x00"]
        }
        foreach x {devmajor devminor} {
            set $x [format %d 0[string trim [set $x] " \x00"]]
        }
    } elseif {$magic == "ustar\x00"} {
        # posix tar
        foreach x {uname gname prefix} {
            set $x [string trim [set $x] "\x00"]
        }
        foreach x {devmajor devminor} {
            set $x [format %d 0[string trim [set $x] " \x00"]]
        }
    } else {
        # old style tar
        foreach x {uname gname devmajor devminor prefix} { set $x {} }
        if {$type == ""} {
            if {[string match */ $name]} {
                set type 5
            } else {
                set type 0
            }
        }
    }

    return [list name $name mode $mode uid $uid gid $gid size $size mtime $mtime \
                 cksum $cksum type $type linkname $linkname magic $magic \
                 version $version uname $uname gname $gname devmajor $devmajor \
                 devminor $devminor prefix $prefix]
}



#return a stream open for reading at the beginning of the tar archive. 
tarjar::_::eproc ::tarjar::connect {tarjar {type ""}} {
    #puts stderr "------>>>  tarjar::connect $tarjar  type:$type"
    upvar #0 ::tarjar::connected con
    
    set original_connectpath $tarjar
    set tarjar [::tarjar::normalize $tarjar] ;#like [file normalize..] except tilde-preserving.
    #set tarjar [string tolower $tarjar] ;#must lowercase after normalize! (why??)

    #Key on lowercased and normalized version of path passed in.
    #puts stdout "%%%tarjar%%%% $tarjar"
    set con(connectpath,$tarjar) $original_connectpath 
    
	
    set is_sourced [expr {[file normalize $tarjar] eq [file normalize [info script]]}]
	
	tarjar::Log debug "'is_sourced':$is_sourced ([file normalize $tarjar] eq [file normalize [info script]])"
	#review - when is it NOT sourced??
	
    if {![string length $tarjar]} {
        error "ERROR: tarjar path required for connection"
    }
  
	if {$tarjar in $con(to)} {
		#todo - count number of connections to particular tarjar and decrement on each disconnect til final?
		#error "ERROR: already connected to $tarjar"
		::tarjar::Puts stderr "WARNING: already connected to $tarjar"
		return [list ok ALREADY_CONNECTED]
	}

	lappend con(to) $tarjar

    #puts "---------->connect $tarjar ---> [info level] [info script]"
    
    set con(location,$tarjar) [file dirname $tarjar]
    set con(startdata,$tarjar) -1
    set con(type,$tarjar) $type
    set con(fh,$tarjar) ""
	
    if {[string range [file tail $tarjar] 0 7] eq "#tarjar-" } {
        #connecting to unwrapped loadscript file.
        set con(type,$tarjar) "unwrapped"

        #use the containing folder's name to determine package & version
        lassign [::split [file tail [file dirname $tarjar]] -] _tarjarMark con(package,$tarjar) con(version,$tarjar)

        set this_pkg_tm_folder      [file dirname [file dirname $tarjar]] ;# subtract the #tarjar-pkg-x folder (i.e 2 folders up from #tarjar-loadscript.tcl)

        
    } else {
        #connecting to .tm, .tj or .tar archive - but may be still be unwrapped version available
        
        #use the tarjar file's name to determine package & version
        lassign [::split [file rootname [file tail $tarjar]] -] con(package,$tarjar) con(version,$tarjar)
        
        set this_pkg_tm_folder      [file dirname $tarjar]
        
        if {$con(type,$tarjar) ne "unwrapped"} {
            #Not directly connecting to unwrapped version -
            # but may still be redirected there
            
            set unwrappedFolder [file join $con(location,$tarjar) #tarjar-$con(package,$tarjar)-$con(version,$tarjar)]
            if {[file exists $unwrappedFolder]} {
                #folder with exact version-match must exist for redirect to 'unwrapped'.
                set con(type,$tarjar) tarjar-redirecting	
            }
    	}
    }
        #puts stdout " +++++ connected(package,$tarjar)  = [set con(package,$tarjar)]"
    
        #=================================================================================================================
        # - examine module paths to determine if we are actually a 'namespaced' package e.g somepath::somepackage-1.0
        #assume the corresponding pkg-version.tm is one folder up.
        if 0 {
        if {![llength $::tarjar::_::last_scanned_tmlist]} {
            #puts stderr "---> populating ::tarjar::_::last_scanned_tmlist"
            set ::tarjar::_::last_scanned_tmlist [tcl::tm::list]
            set ::tarjar::TEMP_auto_path $::auto_path[set ::auto_path {}]
            catch {package require tarjar-invalidname} ;#force tcl to search module paths so that 'package names' is populated
            set ::auto_path $::tarjar::TEMP_auto_path[unset ::tarjar::TEMP_auto_path]
        } else {
            set tmlist [tcl::tm::list]
            if {$::tarjar::_::last_scanned_tmlist ne $tmlist} {
                #puts stderr"(tarjar) last_scanned_tmlist ne tcl::tm::list"
                #puts stderr "    last_scanned_tmlist    ->"
                #puts stderr "    - $::tarjar::_::last_scanned_tmlist"
                #puts stderr "    tcl::tm::list          ->"
                #puts stderr "    - $tmlist"
                
                set ::tarjar::_::last_scanned_tmlist $tmlist 

                set ::tarjar::TEMP_auto_path $::auto_path[set ::auto_path {}]
                catch {package require tarjar-invalidname} ;#force tcl to search module paths so that 'package names' is populated
                set ::auto_path $::tarjar::TEMP_auto_path[unset ::tarjar::TEMP_auto_path]
            }
        }
        }
        
        set unwrapped_tm_file       [file join $this_pkg_tm_folder "[set con(package,$tarjar)]-[set con(version,$tarjar)].tm"]
        set con(tmfile,$tarjar) $unwrapped_tm_file
        #puts stdout "->>> unwrapped_tm_file: $unwrapped_tm_file  tarjar: $tarjar"
        set tail_segments       [list]
        set lcase_tmfile_segments     [string tolower [file split $this_pkg_tm_folder]]
        set lcase_modulepaths         [string tolower [tcl::tm::list]]
        foreach lc_mpath $lcase_modulepaths {
            set mpath_segments [file split $lc_mpath]
            if {[llength [tarjar::structset::intersect $lcase_tmfile_segments $mpath_segments]] == [llength $mpath_segments]} {
                #$lcase_tmfile_segments includes all of $mpath_segments i.e our tm file is within this module path
                set tail_segments [lrange [file split $this_pkg_tm_folder] [llength $mpath_segments] end] ;#use properly cased tail

                #set tail_segments [tarjar::structset::difference $lcase_tmfile_segments $mpath_segments]
                break
            }
        }
        if {[llength $tail_segments]} {
            set con(fullpackage,$tarjar) [join [concat $tail_segments [set con(package,$tarjar)]] ::]  ;#full name of package as used in package require
        } else {
            set con(fullpackage,$tarjar) [set con(package,$tarjar)]
        }
        #puts stdout "\t\t####### Full package name: [set con(fullpackage,$tarjar)]"
        #=================================================================================================================

    
    
    
    
    
    
    
    
    
	
	switch -- $con(type,$tarjar) {
	"tarjar-redirecting" {
		#We're about to redirect to the unwrapped version 
		#Don't want to keep sourcing self 
		# ...but we should keep a connection to ourself open.
		#'returneval' used to abort 'sourcing' of the wrapped version

		
		#connect to the tarball - start at 1st header.
		set con(startdata,$tarjar) 0
		set fh [open $tarjar r] 
		set con(fh,$tarjar) $fh
		fconfigure $fh -encoding binary -translation binary -eofchar {}
		if {![catch {::tarjar::_::readHeader [read $fh 512]}]} {
			seek $fh $con(startdata,$tarjar) start	
			#!todo - efficiency: cache this so first read can use it?
		}

		
		set loadscript_name [file join $unwrappedFolder #tarjar-loadscript-$con(package,$tarjar)-$con(version,$tarjar).tcl]
		if {![file exists   $loadscript_name]} {
			#no match for exact version - take largest version available.
			set versioned_loadscripts [glob -nocomplain -dir $unwrappedFolder -tail #tarjar-loadscript-$con(package,$tarjar)-*.tcl]

			#version numbers cannot contain '-' char!
			if {[llength $versioned_loadscripts]} {
				set max 0
				foreach fname $versioned_loadscripts {		
					set v [lindex [split [file rootname $fname] -] end]
					if {[package vcompare $v $max] == 1} {	
						#$v larger
						set max $v
					}
				}
				set loadscript_name [file join $unwrappedFolder #tarjar-loadscript-$con(package,$tarjar)-$max.tcl]
			} else {
				#try versionless
				set loadscript_name [file join $unwrappedFolder #tarjar-loadscript-$con(package,$tarjar).tcl]
			}
            if {![file exists $loadscript_name]} {
                set loadscript_name [file join $unwrappedFolder #tarjar-loadscript.tcl]
            }
		}

		
		if {$is_sourced} {
			#disconnect from the .tm jar prior to sourcing unwrapped version.
			tarjar::disconnect $tarjar

			#::tarjar::Puts stderr "---> about to source unwrapped version $loadscript_name"
			set script "uplevel #0 ::source -encoding utf-8 $loadscript_name"
            #!todo - http://www.tcl.tk/cgi-bin/tct/tip/90.html  instead of 'returneval'


			set prev_info_script [info script]

			#puts stderr "===> about to source $loadscript_name"
			if {[catch "info script [list $loadscript_name];$script" errm]} {
				#info script {}
				info script $prev_info_script
				::tarjar::_::returneval "error [list [set ::errorInfo]\n]"
			} else {
				::tarjar::Puts stderr "===> stopping .tm sourcing action"
				#info script {}
				info script $prev_info_script
				::tarjar::_::returneval "return [list ok -redirection $loadscript_name]"
			}
		} else {
            #? comments required. When do we get here?
			#error "WTF?"
            puts stderr "|debug> Connecting to '$tarjar' without sourcing - manual connect? Will redirect to '$loadscript_name'"
			tarjar::disconnect $tarjar
			return [tarjar::connect $loadscript_name]
		}
	}	
	"unwrapped" {
        if {[string length [info commands ::thread::id]]} {
            set from [pid],[thread::id]
        } else {
            set from [pid]
        }
        ::tarjar::Puts stderr "$from-> Package $con(package,$tarjar)-$con(version,$tarjar) is using unwrapped version: $tarjar"
        return [list ok ""]
	}
	default {
		# "tarjar"
		#connect to the tarball - start at 1st header.
		set con(startdata,$tarjar) 0

		set fh [open $tarjar r] 
		set con(fh,$tarjar) $fh
		fconfigure $fh -encoding binary -translation binary -eofchar {}
	}
	}

	if {$con(startdata,$tarjar) >= 0} {
		#verify we have a valid tar header

		if {![catch {::tarjar::_::readHeader [read $fh 512]}]} {
			seek $fh $con(startdata,$tarjar) start	
			#!todo - efficiency: cache this so first read can use it?
			return [list ok $fh]	
		} else {
			# this is just a test - probably shouldn't raise an error here
			#error "cannot verify tar header"
		}
	}

	
	set con(to) [lrange $con(to) 0 end-1]
	set con(startdata,$tarjar) -1
	unset con(fh,$tarjar)
	close $fh
	return [list err {Does not appear to be a valid tarjar}]
}
proc ::tarjar::disconnect {{tarjar ""}} {
	variable connected
	tarjar::Log debug " disconnecting $tarjar"



	if {![llength $connected(to)]} {
		return 0
	}
	if {$tarjar eq ""} {
		::tarjar::Puts stderr "WARNING: tarjar not explicitly specified for tarjar::disconnect. Disconnecting last connected: [lindex $connected(to) end]\n - It is better to explicitly specify the tarjar. e.g 'tarjar::disconnect \[info script\]'"
		set tarjar [lindex $connected(to) end]
	}

	if {[set posn [lsearch $connected(to) $tarjar]] == -1} {
		::tarjar::Puts stderr "WARNING: disconnect called when not connected: $tarjar"
		return 0
	}


	if {[string length $connected(fh,$tarjar)]} {
		close $connected(fh,$tarjar)
	}

	if {!$::tarjar::debug} {
        array unset connected *,$tarjar
		#array unset connected fh,$tarjar
		#array unset connected type,$tarjar
		#array unset connected startdata,$tarjar
		#array unset connected location,$tarjar
		#array unset connected package,$tarjar
		#array unset connected version,$tarjar
	
		set connected(to) [lreplace $connected(to) $posn $posn]
	} else {
		tarjar::Log debug "tarjar::debug is 1 - 'disconnect' leaving data in 'tarjar::connected' array"
	}
	
	return 1
}



tarjar::connect [info script] ;#give tarjar a chance to connect to the unwrapped version instead


#package provide tarjar [namespace eval tarjar {
#    set version [lindex [split [file rootname [info script]] -] end]
#}]

package provide tarjar [namespace eval tarjar {
	variable version
    set fname [file tail [file rootname [info script]]]

    if {![string match "#tarjar-*" $fname]} {
	    lassign [lrange [split $fname -] end-1 end] pkgname version
    } else {
        lassign [scan [file tail [file rootname [info script]]] {#tarjar-loadscript-%[a-z]-%s}] pkgname version
        if {![string length $version]} {
            #try again using the parent folder name.
            lassign [scan [file tail [file dirname [info script]]] {#tarjar-%[a-z]-%s}] pkgname version
        }
    }
    set version
}]










#The caching mechanism is an ugly hack anyway.. todo - investigate if there is a modern alternative to directly load (vfs?)
proc ::tarjar::check_or_make_cached_sofile {sofile fileinfoarrdata filename frompack} {
    array set fileinfo $fileinfoarrdata

    if {![file exists $sofile]} {
        set cachedir [file dirname $sofile]
        set filedata [lindex [tarjar::get $filename $frompack] 1]
        file mkdir $cachedir

        set fh [open $sofile w]
        fconfigure $fh -translation binary 
        puts -nonewline $fh $filedata
        close $fh

        #'touch' the file to give it the same mtime as the packaged file.
        file mtime $sofile $fileinfo(mtime)
    } else {
        #!todo - check fstat and copy out IFF mismatching size &/or date?
    
        #set c [file size $sofile]
        file stat $sofile cachefile
        set c $cachefile(size)

        #array set fileinfo [lindex [tarjar::getFileInfo $file $frompack] 1]
        set p $fileinfo(size)

        #puts stderr "!!!size cached:$c pack:$p"

        if {($p != $c) || ($fileinfo(mtime) != $cachefile(mtime))} {
            ::tarjar::Puts stderr "warning - file: $filename. Filesize and/or mtime mismatch, cachedsize:$c packsize:$p;  cachedmtime:$cachefile(mtime) packmtime:$fileinfo(mtime). To use this tarjar's version, remove $sofile and retry."
        } else {
            #puts stderr "OK - file: $file - size $p"
        }
    }
}
#load
#changed args to be more consistent with standard Tcl 'load' command.
#old api ::tarjar::load {filename {frompack ""}}
proc ::tarjar::load {filename args} {
	#puts "%%%%%%%%%%%%% load $filename"
	
	set arglist [list]
	set options [list]

	for {set i 0} {$i < [llength $args]} {incr i} {
		set a [lindex $args $i]
		if {[string first - $a] == 0} {
			incr i ;#skip to value
			lappend options $a
			lappend options [lindex $args $i]
		} else {
			#not an option 
			lappend arglist $a
		}
	}

	set packagename [lindex $arglist 0]
	set interp		[lindex $arglist 1]
	

	if {([llength $arglist] % 2) == 0} {
		array set opts $arglist
        
	}
	if {![info exists opts(-from)]} {
		set opts(-from) [info script]
	}	

	#puts "--->>>1 [info script]"
	#puts "--->>>2 [uplevel 1 {info script}]"
	#puts "---XXX3 $frompack"

	variable connected
	#!todo - copy it out to filesystem (and other binary files?) and load it.

	set frompack [tarjar::_::connect_if_not $opts(-from)] ;#ensure we're connected so we can get $frompack
	
	
	lassign [tarjar::getFileInfo $filename $frompack] stat info
	
	if {$stat ne "ok"} {
		error "unable to find file:'$filename' in tarjar:'$frompack'"
	}

	array set fileinfo $info

	if {$connected(type,$frompack) eq "unwrapped"} {
		set cachedir $connected(location,$frompack)
		set sofile [file join $cachedir $filename]
	} else {
        #puts stderr "wrapped load of file '$filename'"
		set cachedir [file join ~ .TclSoCache [file dirname $filename]]
        #puts stderr "cachefolder: $cachedir"
		set sofile [file join $cachedir [file tail $filename]]
        #puts stderr "sofile: $sofile"
        if {[catch {check_or_make_cached_sofile $sofile [array get fileinfo] $filename $frompack} msg]} {
            after 100
            #retry once in case we're fighting another thread :/
            puts stderr "|tarjar::load> retrying load of cache sofile $sofile once"
            flush stderr
            check_or_make_cached_sofile $sofile [array get fileinfo] $filename $frompack
        }
	
	}

    #review
    #why cd to loading folder?? 
    # - danger of leaving working dir in wrong folder at any error point.
    # - what was the rationale for cd before 'load' anyway?
	#temporarily cd to sofile location for loading. 
	#set place [pwd]
	#cd $cachedir
	if {[string length $packagename]} {
		if {[string length $interp]} {
			#uplevel #0 ::load $sofile $packagename $interp
			set loadcmd [list uplevel 1 ::load $sofile $packagename $interp]
		} else {
			#uplevel #0 ::load $sofile $packagename
			set loadcmd [list uplevel 1 ::load $sofile $packagename]
		}
	} else {
		#uplevel #0 ::load $sofile
		set loadcmd [list uplevel 1 ::load $sofile]
	}
    if 0 {
        if {[catch {
            #thread package may not be loaded
            puts stderr "|loading> $sofile in thread [thread::id]"
        }]} {
            puts stderr "|loading> $sofile"
        }
    }
    flush stderr
    if {[catch {{*}$loadcmd} msg]} {
        if {[string match "*socket*" $msg]} {
            #windows wsastart intermittent errors.. retry once?
            puts stderr "|load_error> socket related error loading '$sofile' - $msg - retrying once"
            flush stderr
            ;#synchronous sleep
            after [expr {int(floor(rand() * 150)) + 150}]
            {*}$loadcmd
        } else {
            puts stderr "|load error> failed to load '$sofile' - $msg"
            flush stderr
            after [expr {int(floor(rand() * 50)) + 100}]
            {*}$loadcmd
            #return -code error $msg
        }
    }
	#cd $place
}


#!todo - use #tarjar-xxx folder as base only if 'relative' path specified
#old api proc ::tarjar::source {filename {frompack ""}}

proc ::tarjar::source {filename args} {
	variable connected
	if {([llength $args] % 2) == 0} {
		array set opts $args
	}
	if {![info exists opts(-from)]} {
		set opts(-from) [info script]
	}
	#!todo - what about -rsrc & -rsrcid options to 'source' for Mac platform?
	set frompack [tarjar::_::connect_if_not $opts(-from)] ;#ensure we're connected so we can get $frompack
	lassign [tarjar::get $filename $frompack] stat dat
	#$dat is the raw sourcefile, so we need to convert any \r\n lineendings to \n which Tcl requires to 'eval'
	set dat [string map {\r\n \n} $dat]
    #puts stdout "---XXXX--- filename: $filename  stat: $stat  dat: $dat"
	set prev_info_script [info script]
	if {$stat eq "ok"} {
        #uplevel #0 [list if 1 "info script [list $filename];$dat"]
		# tailcall useful at all here?
        uplevel 1 [list if 1 "info script [list $filename];$dat"]
        #info script {}
        info script $prev_info_script  ;#reset info script to previous value
	} else {
        error "couldn't source file $filename from tarjar $frompack. reason:$dat"
	}
}

proc ::tarjar::_::connect_if_not {frompack} {
	upvar #0 ::tarjar::connected connected
	#puts *---*---*
	#puts "connect_if_not $frompack"
	#parray connected
	#puts *---*---*
	set frompack [::tarjar::normalize $frompack] ;#like [file normalize..] except tilde-preserving.
	set docon 0
	if {![llength $connected(to)]} {
		if {![string length $frompack]} {
			error "Not connected to a tarjar file, and no tarjar specified"
		} else {
			set docon 1			
		}
	} else {
		if {![string length $frompack]} {
			set frompack [lindex $connected(to) end]
			::tarjar::Puts stderr "WARNING: using last connected tarjar: $frompack for operation\n - tarjar not explicitly specified during operation: [info level -1]"
		} else {
            if {$frompack ni $connected(to)} {
                set docon 1
            }
			#if {([lsearch $connected(to) $frompack] == -1) } {
		    #	set docon 1
			#}
		}
	}
	if $docon {
		if {[lindex [tarjar::connect $frompack] 0] ne "ok"} {
			error "file $frompack does not seem to be a valid tarjar archive"
		} else {
			#we're now connected.
			return $frompack
		}
	}
	#We were already connected
	return $frompack
}

proc ::tarjar::getFileInfo {filename {frompack ""}} {

	#puts "%%%%%%%%%%%%% getFileInfo $filename"

	variable connected
	set frompack [::tarjar::_::connect_if_not $frompack]
	set fh $connected(fh,$frompack)

	if {$connected(type,$frompack) eq "unwrapped"} {
		#for unwrapped connection - $connected(location) already points to the #tarjar-pkg-ver folder
		if {[string range $filename 0 0] eq "/"} {
			#absolute path
			set path [file join $connected(location,$frompack) .. [string trim $filename /]]
		} else {
			#relative path - use #tarjar-xxx as base
			set path [file join $connected(location,$frompack) $filename]
		}

		file stat $path info
		return [list ok [array get info]]
	} else {
		set pkg $connected(package,$frompack)
		set ver $connected(version,$frompack)
		while {![eof $fh]} {
			array set header [::tarjar::_::readHeader [read $fh 512]]
			if {$header(name) eq ""} break
			set name [string trimleft $header(prefix)$header(name) /]

			if {($filename eq $name) || ("#tarjar-${pkg}-${ver}/$filename" eq $name)} {	
				#!todo - set access position to header of first unread file instead?
				#!todo - maintain index of headers read so-far? 
				#(what about added complexity when we also write to tarjar?)
				seek $fh $connected(startdata,$frompack) start 
				return [list ok [array get header]]		
			}
			seek $fh [expr {$header(size) + [::tarjar::_::pad $header(size)]}] current
		}
		seek $fh $connected(startdata,$frompack) start 
	}

	return [list err "file $filename not found in archive"]
}


#get the raw file data - ie no lineending translations done during read
proc ::tarjar::get {filename {frompack ""}} {
    #puts "tarjar::get '$filename' from:'$frompack'"
	variable connected
	set frompack [::tarjar::_::connect_if_not $frompack]
	set fh $connected(fh,$frompack)

	if {$connected(type,$frompack) eq "unwrapped"} {
		#for unwrapped connection - $connected(location) already points to the #tarjar-pkg-ver folder
		if {[string range $filename 0 0] eq "/"} {
			#absolute path
			set path [file join $connected(location,$frompack) .. [string trim $filename /]]
		} else {
			#relative path - use #tarjar-xxx as base
			set path [file join $connected(location,$frompack) $filename]
		}
		set fd [open $path r]
		#fconfigure $fd -encoding utf-8
		fconfigure $fd -encoding binary -translation binary
		return [list ok [lindex [list [read $fd [file size $path]]   [close $fd]  ] 0]]
		#set fd [open $path r]
		#set filedata [read $fd [file size $path]]
		#close $fd
		#return [list ok $filedata]
	} else {
		set pkg $connected(package,$frompack)
		set ver $connected(version,$frompack)
		##!test & review. Any issues with 'seek' and utf-8?
		#fconfigure $fh -encoding utf-8
		#2018 - the answer is yes - it causes corruption of .dll files!
		
		fconfigure $fh -encoding binary -translation binary
		
		while {![eof $fh]} {
			array set header [::tarjar::_::readHeader [read $fh 512]]
			::tarjar::_::HandleLongLink $fh header
			if {$header(name) eq ""} break

			set hdrname [string trimleft $header(prefix)$header(name) /]
            #puts stderr "hdrname: $hdrname  vs  #tarjar-${pkg}-${ver}/$filename"
			if {($filename eq $hdrname) || ("#tarjar-${pkg}-${ver}/$filename" eq $hdrname)} {	
				set filedata [read $fh $header(size)]

				seek $fh $connected(startdata,$frompack) start 
				#!todo - set access position to header of first unread file instead?

				return [list ok $filedata]			
			}
			seek $fh [expr {$header(size) + [::tarjar::_::pad $header(size)]}] current
		}
		seek $fh $connected(startdata,$frompack) start 
	}

	return [list err "file $filename not found in archive"]
}
proc ::tarjar::_::HandleLongLink {fh hv} {
    upvar 1 $hv header thelongname thelongname

    # @LongName Part I.
    if {$header(type) == "L"} {
		# Size == Length of name. Read it, and pad to full 512
		# size.  After that is a regular header for the actual
		# file, where we have to insert the name. This is handled
		# by the next iteration and the part II below.
		set thelongname [string trimright [read $fh $header(size)] \000]
		seekorskip $fh [pad $header(size)] current
		return -code continue
    }
    # Not supported yet: type 'K' for LongLink (long symbolic links).

    # @LongName, part II, get data from previous entry, if defined.
    if {[info exists thelongname]} {
		set header(name) $thelongname
		# Prevent leakage to further entries.
		unset thelongname
    }

    return
}



#!todo - implement/test utf-8 filenames
proc ::tarjar::glob {folder {glob *} {frompack ""}} {
	variable connected
	#parray connected
	#puts "*************"

	set frompack [::tarjar::_::connect_if_not $frompack]
	set fh $connected(fh,$frompack)
	#puts ">>>>>glob $folder $glob - whilst connected to $frompack . fh: $fh"

	if {$connected(type,$frompack) eq "unwrapped"} {
		#for unwrapped connection - $connected(location) already points to the #tarjar-pkg-ver folder
		if {[string range $folder 0 0] eq "/"} {
			#absolute path
			set path [file join $connected(location,$frompack) .. [string trim $folder /]]
		} else {
			#relative path - use #tarjar-xxx as base
			set path [file join $connected(location,$frompack)  $folder]
		}
		#puts "--->globbing -nocomplain -directory $path -tail $glob"
		return [list ok [::glob -nocomplain -directory $path -tail $glob]]

	} elseif {$connected(type,$frompack) eq "tarjar-redirectingXXX???"} {
		if {[string range $folder 0 0] eq "/"} {
			#absolute path
			set path [file join $connected(location,$frompack) [string trim $folder /]]
		} else {
			#relative path - use #tarjar-xxx as base
			set path [file join $connected(location,$frompack) #tarjar-$connected(package,$frompack)-$connected(version,$frompack) $folder]
		}

		return [list ok [::glob -nocomplain -directory $path -tail $glob]]
	} else {
		if {[string range $folder 0 0] eq "/"} {
			#absolute path - assume full path was given
			set path [string trim $folder /]
		} else {
			#relative path - use #tarjar-xxx as base
			set path [file join #tarjar-$connected(package,$frompack)-$connected(version,$frompack) $folder]
		}
		#puts "--->doglob $folder $glob - path:$path"

		set matchlist [list]
		while {![eof $fh]} {
			array set header [::tarjar::_::readHeader [read $fh 512]]
			if {$header(name) eq ""} break

			set name [string trimleft $header(prefix)$header(name) /]

			set segs [::split $name /]
			set psegs [::split $path /]

			set dirsegs [lrange $segs 0 [llength $psegs]-1 ] ;#!?
			set testpath [::join $dirsegs /]

			if {$testpath eq "."} {set testpath ""}
	
           # puts "---name: $name ---"
			#puts "***$path vs $testpath ***"
			
			if {$path eq $testpath} {
				#same folder
				set sub [lindex $segs [llength $dirsegs]] ;#next level folder or filename
				if {[string match $glob $sub]} {
					#if {[lsearch $matchlist $sub] == -1} {
                    if {[string length $sub] && ($sub ni $matchlist)} {
						lappend matchlist $sub
                    }
					#}
				}	
			}
			seek $fh [expr {$header(size) + [::tarjar::_::pad $header(size)]}] current
		}
		seek $fh $connected(startdata,$frompack) start 

		#puts "---GLOB--- returning $matchlist"
		return [list ok $matchlist]	
	}

	return [list err "file $filename not found in archive"]
}

#Check if string is a valid version number for Tcl .tm files.
proc ::tarjar::_is_valid_tm_version {v} {
	#todo - something proper. Note that version numbers can contain letters e.g 7.4a0
	if {[string is digit -strict [string map {. ""} $v]]} {
		return true
	} else {
		return false
	}
}








#tarjar::update
proc ::tarjar::update {tarjar files args} {
	package require tar
	set ok [list]
	set alreadyin [tar::contents $tarjar]
	if {[string range [lindex $alreadyin 0] 0 17] eq "#tarjar-loadscript"} {
		set alreadyin [lrange $alreadyin 1 end]
	}
	if {[string range [lindex $files 0] 0 17] eq "#tarjar-loadscript"} {
		set alreadyin [linsert $alreadyin 0 [lindex $files 0]]
		set files [lrange $files 1 end]
	}
	foreach f $files {
		if {[lsearch $alreadyin $f] < 0} {
			lappend ok $f
		}
	}
	lappend args -update 1
	::tarjar::create $tarjar [concat $alreadyin $ok] {*}$args
}





#todo - take scriptfilelist - where each element can be a file glob
proc ::tarjar::wrap {scriptfile args} {
    set opts(-name) "" ;#package name
    set opts(-version) ""
    set opts(-provide) 1
    set opts(-zip) 0
    set opts(-force) 0  ;#-force 1 to insist we try to create the tarjar no matter what the inputfile format, or to allow wrapping a script that already has a package provide statement

  	set result [dict create status unspecified]
    set allowed_opts {-name -version -provide -zip -force}

    dict for {k v } $args {
        if {$k ni $allowed_opts} {
            error "usage: tarjar::wrap scriptfile ?-version \$v? ?-provide 1|0? ?-zip 0|1? ?-force 0|1?"
        }
    }

    if {[file isdirectory $scriptfile]} {
        #!todo - allow this!
        set msg "Sorry - wrapping a folder is not yet implemented"
        puts stderr "|tarjar::wrap err>$msg"
        error $msg
    }


    array set opts $args

    if {[string is digit -strict [string range [file extension $scriptfile] 1 end]]} {
        #file extension numeric.. seems unlikely, so treat filename as extensionless.
        #!todo -review
        Puts stderr "Warning - weird name provided for scriptfile: '$scriptfile' (extension numeric?)"
        set rootname $scriptfile
    } else {
        set rootname [file rootname $scriptfile]
    }


    set tm_version "" 

    set version_from_filename ""
    set filename_parts [split [file tail $rootname] -]
    set possible_versionpart [lindex $filename_parts end]
    ;#rejoin, accounting for possible dash in module name (unsure if really necessary)
    if {[_is_valid_tm_version $possible_versionpart]} {
        set scriptfile_name [join [lrange $filename_parts 0 end-1] -] 
        set version_from_filename $possible_versionpart
    } else {
        set scriptfile_name [join $filename_parts -]
        set version_from_filename ""
        set possible_versionpart "" ;#
    }
    if {[string length $opts(-name)]} {
        set modulename $opts(-name)
    } else {
        set modulename $scriptfile_name
    }
    #puts stdout "modulename: $modulename"


    #tm_version from supplied filename may be overridden by -version option.


  	set $opts(-version) [string trim $opts(-version) .] ;#silently discard leading/trailing dots!

    if {[string length $opts(-version)]} {
        if {![_is_valid_tm_version $opts(-version)]} {
          error "Version number '$opts(-version)' is invalid - must be comprised of digits separated by dots only."
        }

        #assume user supplied version replaces file version

        if {[string length $version_from_filename]} {
          #file has a version number - we are replacing it.
          #puts "-->1) [join [lrange $filename_parts 0 end-1] -]"
          #puts "-->2) $opts(-version)"
          set tm_version $opts(-version)
        } else {
          #unversioned filename - add it.
          set tm_version $opts(-version)
        }
    } else {
        if {![string length $version_from_filename]} {
          # no version from filename or from -version option
          #! Default 1st version 0.1 !
          set tm_version 0.1
        } else {
          set tm_version $version_from_filename
        }
    }
    Puts stdout "modulename: $modulename tm_version: $tm_version"
    #return
    set tarjar [file dirname $rootname]/${modulename}-${tm_version}.tm


    set fd [open $scriptfile r]
    fconfigure $fd -encoding utf-8
    set checkdata [read $fd 20]

  	set isZP 0; set isTM 0

	
    if {[string range $checkdata 0 1] eq "PK"} {
        #!todo - proper check for non-SFX zip format
        #Assume we want to add SFX header but forgot the -zip flag
        set opts(-zip) 1
    }

    if {$opts(-zip)} {
        if {[string range $checkdata 0 1] eq "PK"} {
            #!todo - proper check for non-SFX zip format
            #assume we just want to add the SFX header
            ::tarjar::make_zip_tm $scriptfile $tarjar
            return
        }

        if {[string range $checkdata 0 13] eq "#zipped-tarjar"} {
            if {!$opts(-force)} {
              error "Input file is already a self-extracting zip file"
            }
            set isZP 1
        } elseif {[string range $checkdata 0 6] eq "#tarjar"} {
            #only set isTM 1 if we're going to zip it! 
            set isTM 1
        }
    } else {
        if {[string range $checkdata 0 13] eq "#zipped-tarjar"} {
            error "This file is already marked as a zipped-tarjar."
        }		
        if {([string range $checkdata 0 6] eq "#tarjar") && ([string range [file tail $scriptfile] 0 17] ne "#tarjar-loadscript")} {
            #seems we're trying to create a standard (unzipped) tarjar out of an existing tarjar - require '-force' to do that.
            #review
            if {!$opts(-force)} {
              error "Input file is already a tarjar."
            }
        }
    }


    if {!$isZP && !$isTM} {
        seek $fd 0 start
        set scriptdata [read $fd [file size $scriptfile]]
        close $fd
        set append_tags 0
            
        if {[string range [file tail $scriptfile] 0 17] ne "#tarjar-loadscript"} {
            set hdr "[::tarjar::_::scriptComment]\n[::tarjar::_::scriptConnect]\n"
            append hdr  "[::tarjar::_::scriptCommentEnd]\n\n"
                  
            #if {![regexp [string map [list %m% $modulename] {.*package\s*provide\s*%m%.*}] $scriptdata]} {}
            if {![regexp [string map [list %m% $modulename] {.*package\s*provide\s*}] $scriptdata]} {
                  if {$opts(-provide)} {
                      append hdr  "[::tarjar::_::scriptPackageProvide $modulename]\n"
                  } else {
                      puts stderr "|tarjar::wrap> warning - no package provide statement detected, and wrap was called with '-provide 0' "
                      append hdr  "[::tarjar::_::scriptPackageProvideExample $modulename]\n"
                      set append_tags 1
                  }
            } else {
                #already contains a package provide statement for the module.
                if {$opts(-provide)} {
                    puts stderr "|tarjar::wrap> package provide $modulename statement detected. Remove the package provide statement and let tarjar provide the package and version based on the .tm name, or call wrap with -provide 0 if you want to maintain the name and version within the wrapped script file."
                    set result [dict create status err reason "package provide detected" modulename $modulename moduleversion $tm_version tarjar $tarjar]
                    return $result
                }
                #append the example script
                append hdr  "[::tarjar::_::scriptPackageProvideExample $modulename]\n"
                        
                #insert the #<modulecode> substitution section (because it's commented out in the scriptPackageProvideExample)
                set append_tags 1
            }
            if $append_tags {
                append hdr  "\n"
                append hdr "#<modulecode>\n"
                append hdr "\n"
                append hdr "#</modulecode>\n"
                append hdr "\n"
                append hdr "#<sourcefiles>\n"
                append hdr "\n"
                append hdr "#</sourcefiles>\n"
                append hdr "\n"
                append hdr "#<loadfiles>\n"
                append hdr "\n"
                append hdr "#</loadfiles>\n"
            }
            #append hdr  "[::tarjar::_::scriptCommentEnd]\n\n\n[::tarjar::_::scriptTidy]"
            append hdr	"#This tarjar initially generated using tarjar version [package present tarjar]\n"
            append hdr  "# command        : tarjar::wrap [file tail $scriptfile] $args\n"
            append hdr  "# working folder : [file dirname [file normalize $scriptfile]]\n"
            append hdr	"#-------------------------tarjar footer follows-------------------------\n\n"
        } else {
            set tarjar [string range $tarjar 19 end]
            #assume loadscript already contains 'package require tarjar' & 'connect' call etc.
            set hdr \n
        }
        #find the #<modulecode> section and substitute in the scriptdata (!note! - must only look for *single* leading # 
        #set compoundscript [::tarjar::_::template_tag_fill $hdr modulecode $scriptdata] 
        #set compoundscript [::tarjar::_::template_tag_fill $compoundscript sourcefiles ""]
        
        #The file(s) being wrapped are placed within the $modulename subfolder so that git version control can work with the raw tcl for the module.
        #i.e wrapped files go in "#tarjar-{modulename}-{version}/{modulename}/"
        set f $modulename/$scriptfile_name[file extension $scriptfile]
        set script_sourcer [string map [list %f% $f] {
            tarjar::source %f%	
        }]
        set compoundscript [::tarjar::_::template_tag_fill $hdr modulecode $script_sourcer]
        array set create_args $args
        foreach k {-name -version -provide -zip -force} {
            catch {unset create_args($k)}
        }
        
        #assert: $versionpart now has a value
        if {![string length [string trim $tm_version]]} {
            error "Missing value for 'tm_version'"
        }
        
        #::tarjar::create $tarjar [list #tarjar-${modulename}-${tm_version}/$scriptfile] -script "$compoundscript" {*}$create_args
        #::tarjar::create $tarjar [list #tarjar-${modulename}-${tm_version}/${modulename}/$scriptfile] -script "$compoundscript" {*}$create_args
        ::tarjar::create $tarjar [list $scriptfile] -script "$compoundscript" {*}[array get create_args]
        set result [dict create status ok modulename $modulename moduleversion $tm_version tarjar $tarjar]
        package forget $modulename
        set uncompressed $tarjar
    } else {
        close $fd
        set uncompressed $scriptfile
    }

    if {$opts(-zip)} {
        if {![catch {exec zip -v} errmsg]} {
            #::tarjar::Puts stderr "-exec: zip -9 [file rootname $tarjar].zip $uncompressed"
            exec zip -9 [file rootname $tarjar].zip $uncompressed
            ::tarjar::make_zip_tm [file rootname $tarjar].zip $tarjar
            file delete [file rootname $tarjar].zip
        } else {
            error "zip command appears to be unavailable Try installing info-zip (e.g chocolatey Info-ZIP for windows, archivers/zip pkg for FreeBSD). err: $errmsg"
        }
    }
    return $result
}

proc ::tarjar::_::template_tag_fill {template tag fill} {
    set result ""
    set mode writing_hdr
    set template [string map [list \r\n \n] $template]
    foreach ln [split $template \n] {
        set testln [string trimleft $ln]
        if {[string match "#<${tag}>*" $testln]} {
            set mode in_tag
            append result $ln\n
            append result $fill\n
            continue
        }
        if {[string match "#</${tag}>*" $testln]} {
            set mode writing_tail
            append result $ln\n
            continue
        }
        if {$mode in {"writing_hdr" "writing_tail"}} {
            append result $ln\n
        } else {
            #presumably we are 'in_tag' - ie - ignoring anything found within that block because it is being replaced.
        }
    }
    return $result
}


proc ::tarjar::_::scriptComment {} {
return {
#START-tarjar-loadscript-header----------------------------------------------------------------------------
#START-header-comments
#If there is anything other than whitespace above this header, then this is a tarjar in packed form. 
# (e.g If there is a line above the "START-.." lines)
#
# * Do not edit the code whilst in 'packed' form * 
#
# - untar to edit, and retar it afterward using a standard tar program with the format ustar, gnu or v7 (posix & pax formats will not work)
# - When untarred - there will be a folder named #tarjar-<packagename>-<version>
#   You can then edit the #tarjar-loadscript.tcl and other script files in the folder.
#
#If there is no data above the header, it is an unpacked fragment of a tarjar and may be edited.
#Make sure however that your editor preserves the trailing comment (#) as the final character.
# i.e - no trailing newline or carriage-return.
#
# A tarjar is a valid tar archive in which the first archived file consists of a tcl script
# containing a leading newline and a trailing comment (#) character and line continuation character.
# The comment character and line continuation character hide the tar-header for the next file from Tcl.
# This first file in the tarball must be named with the prefix #tarjar-loadscript 
#
# The next file is named #z and contains a final ctrl-z to tell Tcl it has reached the end of scripts 
#  (and thus to stop interpreting).
# The #z file is separate from the initial script file because some editors may not be able to handle the
# ctrl-z character.
# The tarball should have all its contents within a single directory named #tarjar-<package>-<version>
# - This allows you to safely untar any package 'in place'
# - If a package on your 'tcl module path' is unwrapped (i.e untarred) - then the unwrapped files
#   are the ones that will be loaded by the interpreter.
#
# This header (excluding this comment block) and the call to tarjar::disconnect are needed to:
# a) allow 'package require' to redirect to the unwrapped version of the tarjar
# b) enable sourcing & loading of any other files contained in the tarjar
#END-header-comments
}
}

proc ::tarjar::_::scriptConnect {} {
return {
if {[catch {package present tarjar}]} {
    set ::TEMP_auto_path $::auto_path; set ::auto_path [list]
    if {![catch {package require tarjar} msg]} {
        set ::auto_path $::TEMP_auto_path; unset ::TEMP_auto_path
        #Do not wrap 'tarjar::connect' in its own 'catch'!
        #for unwrapped execution, tarjar::connect may need to abort the 'source' operation.
        ::tarjar::connect [info script]
    } else {
        puts stderr "|tarjar> error during package require tarjar '$msg'"
        set ::auto_path $::TEMP_auto_path; unset ::TEMP_auto_path
    }
} else {
    tarjar::connect [info script]   
}
}
}


proc ::tarjar::_::scriptPackageProvideExampleXXXX {modulename} {
return [string map [list %m% $modulename] {

##example 'package provide' statement.
#package provide %m% [namespace eval %m% {
#	variable version
#
#	#set version [lindex [split [file rootname [info script]] -] end]
#	set version [tarjar::myversion]
#}]





}]
}

proc ::tarjar::_::scriptPackageProvideXXX {modulename} {
    if {![string length $modulename]} {
        error "missing modulename"
    }
    
    return [string map [list %m% $modulename] {

package provide %m% [namespace eval %m% {
    variable version
    set version [lindex [split [file rootname [info script]] -] end]
}]

    }]
}

proc ::tarjar::_::scriptPackageProvideExample {args} {
    set script [scriptPackageProvide $args]
    
    set egscript    "##example 'package provide' block.\n"
    append egscript "## - unnamespaced procs placed within #<modulecode/> section will produce commands within a namespace matching the module name.\n"
    foreach ln [split $script \n] {
        append egscript "#$ln\n"
    }
    return $egscript
}


proc ::tarjar::_::scriptPackageProvide {args} {
    #note: the commented <modulecode> etc section below is not just a comment.
    # - these tags are expected to exist. 
	
	if {[string tolower [file extension $::tarjar::connected(self)]] eq ".tm"} {
		package require tar
		set snippet [tar::get $::tarjar::connected(self) #tarjar-tarjar-${::tarjar::version}/codesnippets/full_loader.txt]
	} else {
		set snipfile $::tarjar::connected(self)/codesnippets/full_loader.txt
		set fd [open $snipfile r]
		#for consistency, load our own source files as utf-8 too
		fconfigure $fd -encoding utf-8
		set snippet [read $fd [file size $snipfile]]
		close $fd
	}
	return $snippet
}











proc ::tarjar::_::scriptCommentEnd {} {
return {
#
#END-tarjar-loadscript-header------------------------------------------------------------------------------
}
}

proc ::tarjar::_::scriptTidy {} {
return {
#START-tarjar-loadscript-tidy------------------------------------------------------------------------------
#
if {![catch {package require tarjar}]} {
    ::tarjar::disconnect [info script]
}
#
#END-tarjar-loadscript-tidy--------------------------------------------------------------------------------
}
}







namespace eval ::tarjar::man {
    proc repack {tarjar args} {
        puts stdout ""
        tarjar::create $tarjar {} {*}$args
    }


    proc ls {tarjar args} {
        package require tar
        tar::contents $tarjar
    }


}




#tarjar::create 
# tarjar : output file name (the 'standard' extension is .tm but this is not enforced) 
# files   : list of files & directories to be archived.
#
## optional arguments
#-script : code for load-script. Ignored for tarjar if #tarjar-loadscript specified as 1st file in archive.
#-tarball : if 1 create valid tarball (tarjar) else create a hybrid script & tarball (tarbaby)
proc ::tarjar::create {tarjar files args} {
	#Puts stdout "called: 'create $tarjar $files $args'"
	package require tar

	variable loadables
	variable sourceables

    set allowed_opts {-script -modulecode -update -version -provide -zip -force}

    dict for {k v} $args {
        if {$k ni $allowed_opts} {
            error "usage: tarjar::create tmfilename filelist -script \$complete_loadscript -modulecode \$procs -update 0|1\n received '$args'"
        }
    }
    
    
	set defaults {-script "" -update 0} ;#defaults
	set opts [dict merge $defaults $args]
    

	set script  [dict get $opts -script]
	set update  [dict get $opts -update]
	
	set name_dash_version [file rootname [file tail $tarjar]]
	
	set parts [split $name_dash_version -]
	set modulename [join [lrange $parts 0 end-1] -]
	set moduleversion [lindex $parts end]


	

	#set magicfile "#tarjar-loadscript-$name_dash_version.tcl"
    set magicfile "#tarjar-loadscript.tcl"  ;#we no longer require filename and version here -  unversioned, unnamed magicfile is now the default.

	set mainfolder "#tarjar-$name_dash_version"
	file mkdir $mainfolder
	file mkdir $mainfolder/$modulename
	file mkdir $mainfolder/$modulename/load
	
	#Review! why are we assuming the magicfile is the first in the list?
		#if {[string range [lindex $files 0] 0 17] eq "#tarjar-loadscript" } {}

    if {"#tarjar-loadscript.tcl" in $files} {
        set script_posn [lsearch $files "#tarjar-loadscript.tcl"]
				set magicfile [lindex $files $script_posn]
				#We'll assume no -script argument supplied as it doesn't make sense to specify both.
				set fd [open $magicfile r]
				fconfigure $fd -encoding utf-8
				set script [read $fd [file size $magicfile]]	
				close $fd
        set files [lreplace $files $script_posn $script_posn][set files {}]
		} else {
			if {[string length $script]} {
          if {[file exists $magicfile]} {
            #script specified.. but magicfile (in tarjar destination folder) exists - move it aside. 
            file rename -force $magicfile $magicfile.old
            Puts stderr "$magicfile moved to $magicfile.old"
          }
			} else {
        #!todo - review this branch
				set fn [file join $mainfolder $magicfile]
				if {[file exists $fn]} {
            set fd [open $fn r]
            fconfigure $fd -encoding utf-8
            set script [read $fd [file size $fn]]	
            close $fd
				} else {
					#fallback to plain #tarjar-loadscript if it exists
					#but still insert it as #tarjar-loadscript-<pkg-ver>
					set fn [file join $mainfolder "#tarjar-loadscript"]
					if {[file exists $fn]} {
              set fd [open $fn r]
              fconfigure $fd -encoding utf-8
              set script [read $fd [file size $fn]]	
              close $fd
					}
				}
			}
		}

	if {![string length $script]} {
		#basic script.
		#assume we need to source all .tcl files and load anything else in loadables extension list.
		set script		[::tarjar::_::scriptComment]
		append script	[::tarjar::_::scriptConnect]
		append script	[::tarjar::_::scriptCommentEnd]
        
		#unless we grep all source files for 'package provide $modulename' -
		# then we can only really put in a commented-out sample 'package provide' statement.
		# Also - loaded binaries may make their own package provide call.
		append script	[::tarjar::_::scriptPackageProvide $modulename]

        set script_sourcefiles ""
        set script_loadfiles ""
		append script_sourcefiles {
# - source all .tcl files
}
       append script_loadfiles {
# - load files with known loadable extensions
}

		append script_sourcefiles \n
		foreach fn $files {
			set ext [file extension $fn]
			if {$ext in $sourceables} {
				#.tcl / .tk files
				append script_sourcefiles "::tarjar::source $fn\n"
			} else {
				if {$ext in $loadables} {
					#.dll, .so etc
					append script_loadfiles "::tarjar::load $fn\n"
				}
			}
		}

		foreach fn [::glob -nocomplain -dir $mainfolder -type f *] {
			set ext [file extension $fn]
			if {$ext in $sourceables} {
				append script_sourcefiles "::tarjar::source $fn\n"
			} else {
			  	if {$ext in $loadables} {
					append script_loadfiles "::tarjar::load $fn\n"
				}
			}
			#!todo - also source & load files under platform-specific folders
		}

        if {[dict exists $opts -modulecode]} {
            set script [::tarjar::_::template_tag_fill $script modulecode [dict get $opts -modulecode]]    
        }
        
        set script [::tarjar::_::template_tag_fill $script sourcefiles $script_sourcefiles]    
        set script [::tarjar::_::template_tag_fill $script loadfiles $script_loadfiles]    

        

	}

	set ftemp $files
	set files [list]
	foreach f $ftemp {
		if {![file exists $f]} {
			Puts stderr "warning: file/folder not found - unable to add/update: $f (destination: $destination)"
		} else {
		
		set f_name_dash_version [file rootname [file tail $f]]
	
			set f_parts [split $f_name_dash_version -]
			set f_modulename [join [lrange $f_parts 0 end-1] -]
			set f_moduleversion [lindex $f_parts end]
			
			if {($f_modulename eq $modulename) && ($f_moduleversion eq $moduleversion)} {
				set f_target $f_modulename ;#drop the version number so file suitable for git version control
			} else {
				set f_target $f_name_dash_version ;#unknown - leave it alone.
			}
		
			lappend files $f
			file copy $f $mainfolder/$modulename/${f_target}[file extension $f]
		}
	}



	if {[string range $script 0 0] ne "\n"} {
		set script \n$script
	}
	
        if {[string range $script end-1 end] ne "#\\"} {
            if {[string first "#Do not remove the trailing" $script] == -1} {
                append script {
#Do not remove the trailing 2 characters from this file. Must be a tcl comment char (hash) followed by a line continuation char (backslash) with no following EOL chars.
}
            }
            #trailing comment + line-continuation character - hides next tar header from tcl
            append script "#\\"
        }

        set fd [open [file join $mainfolder $magicfile] w]
        fconfigure $fd -encoding binary -translation binary
        puts -nonewline $fd $script
        close $fd

		#separate ctrl-z file
		if {![file exists [file join $mainfolder #z]]} {
            set fd [open [file join $mainfolder #z] w] 
            fconfigure $fd -encoding binary -translation binary
            puts -nonewline $fd "#This 2-line file marks the end of the Tcl-sourceable section of a tarjar. Do not remove the trailing ctrl-z character from this file.\n"
            puts -nonewline $fd "\u001A"
            close $fd	
		}

		#if {![file exists [file join $mainfolder README.tarjar.txt]]} {
		#!todo	
		#}
		
		
		
		if {$update} {
			set missing [list]
			foreach f [lrange [tar::contents $tarjar] 1 end] {
				#strip leading "#tarjar-modulename-version/modulename/"
				set target_folder "#tarjar-${modulename}-${moduleversion}/$modulename/"
				set tail [string range $f [string length $target_folder] end]
				if {![file exists [file join [pwd] $tail]]} {
					lappend missing $f
				}
			}

			if {[llength $missing]} {
				#missing at least 1 file		
				#we don't have an update feature in the plain tar library.. only create & add
				# so to 'update' we'd have to untar and retar if not all the tarred files are present where we are.
				#for now.. we'll remove missing files from the archive
				#!todo - review behaviour - this possibly violates the 'least surprise' principle.


				puts stdout "Missing [llength $missing] items. They will be *removed* from the archive if you continue. Do you wish to continue? \[Y/N\]"
				set answer [gets stdin]

				if {[string tolower $answer] eq "y"} {
					set remaining [list]
					foreach f $files {
						if {[lsearch $missing $f] < 0} {
							lappend remaining $f
						}					 
					}

					tar::create $tarjar [linsert $remaining 0 $mainfolder]
				}
			} else {
				#everything present.. so we can just recreate the tarjar in place
				tar::create $tarjar [linsert $files 0 $mainfolder]
			}
		} else {
			#puts stdout "about to call 'tar::create $tarjar [linsert $files 0 $mainfolder]'"
			#tar::create $tarjar [linsert $files 0 $mainfolder]
			
			#In general we should only pass the main folder (#tarjar-modulename-moduleversion) to tar::create. 
			#  We don't want other files in the root, or the .tm won't source properly.
			#Puts stdout "about to call 'tar::create $tarjar [list $mainfolder]'"
			tar::create $tarjar [list $mainfolder]  
		}	
}
proc ::tarjar::getTempDir {} {
	if {[info exists ::env(TEMP)]} {
		return $::env(TEMP)
	} elseif {[info exists ::env(TMP)]} {
		return $::env(TMP)
	} elseif {[info exists ::env(TMPDIR)]} {
		return $::env(TMPDIR)
	} elseif {[info exists ::env(TRASH_FOLDER)]} {
		#for macintosh
		return $::env(TRASH_FOLDER)
	} else {
		if {[file exists /tmp]} {
			return /tmp
		} else {
			return [pwd]
		}
	}
}
#maintenance-note:  platform 1.0.17
#used by full_loader.txt
namespace eval ::tarjar::platform {
    #borrow from tcl's platform 1.0.17
    # -- platform::generic
    #
    # Assembles an identifier for the generic platform. It leaves out
    # details like kernel version, libc version, etc.
    proc generic {} {
        global tcl_platform

        set plat [string tolower [lindex $tcl_platform(os) 0]]
        set cpu  $tcl_platform(machine)

        switch -glob -- $cpu {
        sun4* {
            set cpu sparc
        }
        intel -
        ia32* -
        i*86* {
            set cpu ix86
        }
        x86_64 {
            if {$tcl_platform(wordSize) == 4} {
            # See Example <1> at the top of this file.
            set cpu ix86
            }
        }
        ppc -
        "Power*" {
            set cpu powerpc
        }
        "arm*" {
            set cpu arm
        }
        ia64 {
            if {$tcl_platform(wordSize) == 4} {
            append cpu _32
            }
        }
        }

        switch -glob -- $plat {
        windows {
            if {$tcl_platform(platform) == "unix"} {
            set plat cygwin
            } else {
            set plat win32
            }
            if {$cpu eq "amd64"} {
            # Do not check wordSize, win32-x64 is an IL32P64 platform.
            set cpu x86_64
            }
        }
        sunos {
            set plat solaris
            if {[string match "ix86" $cpu]} {
            if {$tcl_platform(wordSize) == 8} {
                set cpu x86_64
            }
            } elseif {![string match "ia64*" $cpu]} {
            # sparc
            if {$tcl_platform(wordSize) == 8} {
                append cpu 64
            }
            }
        }
        darwin {
            set plat macosx
            # Correctly identify the cpu when running as a 64bit
            # process on a machine with a 32bit kernel
            if {$cpu eq "ix86"} {
            if {$tcl_platform(wordSize) == 8} {
                set cpu x86_64
            }
            }
        }
        aix {
            set cpu powerpc
            if {$tcl_platform(wordSize) == 8} {
            append cpu 64
            }
        }
        hp-ux {
            set plat hpux
            if {![string match "ia64*" $cpu]} {
            set cpu parisc
            if {$tcl_platform(wordSize) == 8} {
                append cpu 64
            }
            }
        }
        osf1 {
            set plat tru64
        }
        default {
            set plat [lindex [split $plat _-] 0]
        }
        }

        return "${plat}-${cpu}"
    }

    # -- platform::identify
    #
    # Assembles an identifier for the exact platform, by extending the
    # generic identifier. I.e. it adds in details like kernel version,
    # libc version, etc., if they are relevant for the loading of
    # packages on the platform.

    proc identify {} {
        global tcl_platform

        set id [generic]
        regexp {^([^-]+)-([^-]+)$} $id -> plat cpu

        switch -- $plat {
        solaris {
            regsub {^5} $tcl_platform(osVersion) 2 text
            append plat $text
            return "${plat}-${cpu}"
        }
        macosx {
            set major [lindex [split $tcl_platform(osVersion) .] 0]
            if {$major > 19} {
            set minor [lindex [split $tcl_platform(osVersion) .] 1]
            incr major -9
            append plat $major.[expr {$minor - 1}]
            } else {
            incr major -4
            append plat 10.$major
            return "${plat}-${cpu}"
            }
            return "${plat}-${cpu}"
        }
        linux {
            # Look for the libc*.so and determine its version
            # (libc5/6, libc6 further glibc 2.X)

            set v unknown

            # Determine in which directory to look. /lib, or /lib64.
            # For that we use the tcl_platform(wordSize).
            #
            # We could use the 'cpu' info, per the equivalence below,
            # that however would be restricted to intel. And this may
            # be a arm, mips, etc. system. The wordsize is more
            # fundamental.
            #
            # ix86   <=> (wordSize == 4) <=> 32 bit ==> /lib
            # x86_64 <=> (wordSize == 8) <=> 64 bit ==> /lib64
            #
            # Do not look into /lib64 even if present, if the cpu
            # doesn't fit.

            # TODO: Determine the prefixes (i386, x86_64, ...) for
            # other cpus.  The path after the generic one is utterly
            # specific to intel right now.  Ok, on Ubuntu, possibly
            # other Debian systems we may apparently be able to query
            # the necessary CPU code. If we can't we simply use the
            # hardwired fallback.

            switch -exact -- $tcl_platform(wordSize) {
            4 {
                lappend bases /lib
                if {[catch {
                exec dpkg-architecture -qDEB_HOST_MULTIARCH
                } res]} {
                lappend bases /lib/i386-linux-gnu
                } else {
                # dpkg-arch returns the full tripled, not just cpu.
                lappend bases /lib/$res
                }
            }
            8 {
                lappend bases /lib64
                if {[catch {
                exec dpkg-architecture -qDEB_HOST_MULTIARCH
                } res]} {
                lappend bases /lib/x86_64-linux-gnu
                } else {
                # dpkg-arch returns the full tripled, not just cpu.
                lappend bases /lib/$res
                }
            }
            default {
                return -code error "Bad wordSize $tcl_platform(wordSize), expected 4 or 8"
            }
            }

            foreach base $bases {
            if {[LibcVersion $base -> v]} break
            }

            append plat -$v
            return "${plat}-${cpu}"
        }
        }

        return $id
    }

}
#!todo - make obsolete.  'platform' package is now part of Tcl core - use platform::identify instead.
proc ::tarjar::platform {} {
    set plat [lindex $::tcl_platform(os) 0]
    set mach $::tcl_platform(machine)
    switch -glob -- $mach {
        intel -
        i*86*	{ set mach x86 }
        sun4*	{ set mach sparc }
        Power*	{ set mach ppc }
	  9000*	{ set mach 9000} 
	  default	{ set mach [string map {" " - / -} $mach]}
    }
    return "$plat-$mach"
}

#------------------------------------------------------------------------------------
# zipped-tarjar 
# Original zipped-module creation code by Pascal Scheffers at 'Using a zip file as a Tcl Module' http://wiki.tcl.tk/13869
#------------------------------------------------------------------------------------
	# [make_zip_tm /zipfile/ /outfile/]
	# Prefixes the specified zipfile with the tclmodule mounter stub and writes out 'outfile'

	# [make_sfx_zip /zipfile/ /outfile/ /sfxstub/]
	# Adds an arbitrary 'sfx' to a zip file, and adjusts the central directory
	# and file items to compensate for this extra data.

proc ::tarjar::make_zip_tm { zipfile outfile } {
	set sfx_stub {#zipped-tarjar
#Do not edit this script directly whilst it is part of a zipped .tm file!
# (zip headers need to be adjusted if this script modified)

#foreach {pkg ver} [split [file rootname [file tail [info script]]] -] {break}
lassign [split [file rootname [file tail [info script]]] -] pkg ver

if {[catch {package require vfs::zip}]} {
    error "You need the vfs::zip package to use the $pkg package as is. Alternatively, unzip the module file [info script] and try again."
}

vfs::zip::Mount [info script] [info script]

set tarjar [file join [info script] $pkg-$ver.tm]
if {[file exists $tarjar]} {
    source $tarjar
} else {
	set tarjar_loadscript [file join [info script] #tarjar-$pkg-$ver #tarjar-loadscript-$pkg-$ver]
	::tarjar::Puts stderr "2===> $tarjar not found. Will try to source $tarjar_loadscript"
	if {[file exists $tarjar_loadscript]} {
		source $tarjar_loadscript
	}
}
}
    append sfx_stub \x1A
    tarjar::_::make_sfx_zip $zipfile $outfile $sfx_stub
}


proc tarjar::_::make_sfx_zip { zipfile outfile sfx_stub } {

        set in [open $zipfile r]
        fconfigure $in -translation binary -encoding binary
		#set in_data [read $in [file size $zipfile]]

        set out [open $outfile w+]
        fconfigure $out -translation binary -encoding binary

        puts -nonewline $out $sfx_stub

        set offset [tell $out]

        lappend report "sfx stub size: $offset"

        fcopy $in $out
        close $in

        set size [tell $out]

        # Now seek in $out to find the end of directory signature:
        # The structure itself is 24 bytes long, followed by a maximum of 64Kbytes text

        if { $size < 65559 } {
            set seek 0
        } else {
            set seek [expr { $size - 65559 } ]
        }
        #flush $out
        seek $out $seek
        #puts "$seek [tell $out]"

        set data [read $out]
        set start_of_end [string last "\x50\x4b\x05\x06" $data]

        set start_of_end [expr {$start_of_end + $seek}]
        lappend report "SEO: $start_of_end ([expr {$start_of_end-$size}]) [string length $data]"

        seek $out $start_of_end
        set end_of_ctrl_dir [read $out]

        binary scan $end_of_ctrl_dir issssiis eocd(signature) eocd(disknbr) eocd(ctrldirdisk) \
            eocd(numondisk) eocd(totalnum) eocd(dirsize) eocd(diroffset) eocd(comment_len)

        lappend report "End of central directory: [array get eocd]"

        seek $out [expr {$start_of_end+16}]

        #adjust offset of start of central directory by the length of our sfx stub
        puts -nonewline $out [binary format i [expr {$eocd(diroffset)+$offset}]]
        flush $out

        seek $out $start_of_end
        set end_of_ctrl_dir [read $out]
        binary scan $end_of_ctrl_dir issssiis eocd(signature) eocd(disknbr) eocd(ctrldirdisk) \
            eocd(numondisk) eocd(totalnum) eocd(dirsize) eocd(diroffset) eocd(comment_len)

        lappend report "New dir offset: $eocd(diroffset)"
        lappend report "Adjusting $eocd(totalnum) zip file items."

        seek $out $eocd(diroffset)
        for {set i 0} {$i <$eocd(totalnum)} {incr i} {
            set current_file [tell $out]
            set fileheader [read $out 46]
            binary scan $fileheader is2sss2ii2s3ssii x(sig) x(version) x(flags) x(method) \
                x(date) x(crc32) x(sizes) x(lengths) x(diskno) x(iattr) x(eattr) x(offset)

            if { $x(sig) != 33639248 } {
                error "Bad file header signature at item $i: $x(sig)"
            }

            foreach size $x(lengths) var {filename extrafield comment} {
                if { $size > 0 } {
                    set x($var) [read $out $size]
                } else {
                    set x($var) ""
                }
            }
            set next_file [tell $out]
            lappend report "file $i: $x(offset) $x(sizes) $x(filename)"

            seek $out [expr {$current_file+42}]
            puts -nonewline $out [binary format i [expr {$x(offset)+$offset}]]

            #verify:
            flush $out
            seek $out $current_file
            set fileheader [read $out 46]
            lappend report "old $x(offset) + $offset"
            binary scan $fileheader is2sss2ii2s3ssii x(sig) x(version) x(flags) x(method) \
                x(date) x(crc32) x(sizes) x(lengths) x(diskno) x(iattr) x(eattr) x(offset)
            lappend report "new $x(offset)"

            seek $out $next_file
        }

		close $out
        #puts [join $report \n]
}



#------------------------------------------------------------------------------------
tarjar::disconnect [info script]


#do not remove trailing hash-backslash. There must be no linefeed/carriagereturn etc following it either.
#\
